/**
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 * @format
 */

import {default as generate} from '@babel/generator';
import {parse} from '@babel/parser';
import {transformFromAstSync} from '@babel/core';

function transform({
  filename,
  options,
  src,
}: {
  filename: string;
  config: any;
  options: any;
  src: string;
}) {
  const isPlugin =
    options.projectRoot && options.projectRoot.includes('/desktop/plugins/');
  const isMain =
    options.projectRoot &&
    options.projectRoot.endsWith('/static') &&
    !options.isTestRunner;
  const presets = [
    isMain
      ? [
          require('@babel/preset-env'),
          {targets: {electron: process.env.FLIPPER_ELECTRON_VERSION}},
        ]
      : require('@babel/preset-react'),
  ];
  const isTypeScript = filename.endsWith('.tsx') || filename.endsWith('.ts');
  const ast = parse(src, {
    sourceFilename: filename,
    plugins: isTypeScript
      ? [
          'jsx',
          'typescript',
          'classProperties',
          'optionalChaining',
          'nullishCoalescingOperator',
        ]
      : [
          'jsx',
          ['flow', {all: true}],
          'classProperties',
          'objectRestSpread',
          'optionalChaining',
          'nullishCoalescingOperator',
        ],
    sourceType: 'module',
  });

  // run babel
  const plugins = [];

  if (!isTypeScript) {
    plugins.push(
      require('@babel/plugin-transform-modules-commonjs'),
      require('@babel/plugin-proposal-object-rest-spread'),
      require('@babel/plugin-proposal-class-properties'),
      require('@babel/plugin-transform-flow-strip-types'),
      require('@babel/plugin-proposal-optional-chaining'),
      require('@babel/plugin-proposal-nullish-coalescing-operator'),
      require('./dynamic-requires'),
    );
  } else {
    plugins.push(
      require('@babel/plugin-transform-typescript'),
      require('@babel/plugin-proposal-class-properties'),
      require('@babel/plugin-transform-modules-commonjs'),
      require('@babel/plugin-proposal-optional-chaining'),
      require('@babel/plugin-proposal-nullish-coalescing-operator'),
    );
  }
  if (process.env.FLIPPER_FB) {
    plugins.push(require('./fb-stubs'));
  }
  if (process.env.BUILD_HEADLESS) {
    plugins.push(require('./electron-stubs'));
  }
  if (!options.isTestRunner) {
    if (isMain) {
      // For the main Electron process ("static" folder), to avoid issues with
      // native node modules, we prevent Metro from resolving any installed modules.
      // Instead all of them are just resolved from "node_modules" as usual.
      plugins.push(require('./electron-requires-main'));
      // Metro bundler messes up "global.process", so we're changing all its occurrences to "global.electronProcess" instead.
      // https://github.com/facebook/metro/blob/7e6b3114fc4a9b07a8c0dd3797b1e0c55a4c32ad/packages/metro/src/lib/getPreludeCode.js#L24
      plugins.push(require('./electron-process'));
    } else {
      // Replacing require statements with electronRequire to prevent metro from
      // resolving them. electronRequire are resolved during runtime by electron.
      // As the tests are not bundled by metro and run in @jest-runner/electron,
      // electron imports are working out of the box.
      plugins.push(require('./electron-requires'));
    }
  }
  if (isPlugin) {
    plugins.push(require('./flipper-requires'));
  } else {
    plugins.push(require('./import-react'));
  }
  const transformed = transformFromAstSync(ast, src, {
    ast: true,
    babelrc: !filename.includes('node_modules'),
    code: false,
    comments: false,
    compact: false,
    root: options.projectRoot,
    filename,
    plugins,
    presets,
    sourceMaps: true,
    retainLines: !!options.isTestRunner,
  });
  if (!transformed) {
    throw new Error('Failed to transform');
  }
  const result = generate(
    transformed.ast!,
    {
      filename,
      sourceFileName: filename,
      sourceMaps: true,
      retainLines: !!options.isTestRunner,
    },
    src,
  );
  return {
    ast: transformed.ast,
    code: result.code,
    filename,
    map: result.map,
  };
}

module.exports = {
  transform,
  // Disable caching of babel transforms all together. We haven't found a good
  // way to cache our transforms, as they rely on side effects like env vars or
  // the existence of folders in the file system.
  getCacheKey: () => Math.random().toString(36),
  process(src: string, filename: string, config: any, options: any) {
    return transform({
      src,
      filename,
      config,
      options: {...options, isTestRunner: true},
    });
  },
};
